# -*- python -*-
# $RCSfile: pinnodes.py,v $
# $Revision: 1.32 $
# $Author: langer $
# $Date: 2005/05/13 18:44:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.SWIG.common import switchboard
from oof2.common import debug
from oof2.common import primitives
from oof2.common import toolbox
from oof2.common.IO import oofmenu
from oof2.common.IO import xmlmenudump

# Toolbox for pinning nodes.  Pinned nodes don't move during mesh
# modification operations.  They *can* move during equilibration --
# they're not boundary conditions.  Pinned nodes are treated much like
# Skeleton Selectables, and share a lot of code with them.

class PinnedNodesToolbox(toolbox.Toolbox):
    def __init__(self, gfxwindow):
        toolbox.Toolbox.__init__(self, 'Pin_Nodes', gfxwindow)
        self.pinnednodes = None         # PinnedNodeSelection object
        self.sb_callbacks = [
            switchboard.requestCallback((self.gfxwindow(), 'layers changed'),
                                        self.refresh),
            switchboard.requestCallback(('who changed', 'Skeleton'),
                                        self.skelChanged),
            switchboard.requestCallback("new pinned nodes",
                                        self.newPinnedNodes)
            ]

    def refresh(self):
        skeletoncontext = self.gfxwindow().topwho('Skeleton')
        if skeletoncontext is not None:
            self.pinnednodes = skeletoncontext.pinnednodes # PinnedNodeSelection
        else:
            self.pinnednodes = None

    def skelChanged(self, context):
        if context is self.getSkeletonContext():
            self.refresh()

    def activate(self):
        self.refresh()

    def makeMenu(self, menu):
        self.menu = menu
        menu.addItem(oofmenu.OOFMenuItem(
            'Pin',
            callback = self.pin,
            params=[primitives.PointParameter('point', tip='Target point.')],
            help="Pin the node closest to the given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/pin.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'UnPin',
            callback = self.unpin,
            params=[primitives.PointParameter('point', tip='Target point.')],
            help="Unpin the node closest to the given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/unpin.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'TogglePin',
            callback=self.togglepin,
            params=[primitives.PointParameter('point', tip='Target point.')],
            help="Toggle the pinnedness of the node closest to the given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/toggle_pin.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'UnPinAll',
            callback = self.unpinall,
            help="Unpin all nodes.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/unpin_all.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'Invert',
            callback = self.invert,
            help="Invert pinned nodes.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/invert_pin.xml')
            ))

        menu.addItem(oofmenu.OOFMenuItem(
            'Undo',
            callback=self.undoPin,
            help="Undo the latest pin.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/undo_pin.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'Redo',
            callback=self.redoPin,
            help="Redo the latest undone pin.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/redo_pin.xml')
            ))

    def getSkeletonContext(self):
        if self.pinnednodes is not None:
            return self.pinnednodes.skeletoncontext
        else:
            return None
        
    def getSkeleton(self):
        if self.pinnednodes is not None:
            return self.pinnednodes.skeletoncontext.getObject()

    def pin(self, menuitem, point):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.pinPoint(point)
            self.pinnednodes.signal()

    def unpin(self, menuitem, point):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.unpinPoint(point)
            self.pinnednodes.signal()

    def togglepin(self, menuitem, point):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.togglepinPoint(point)
            self.pinnednodes.signal()

    def invert(self, menuitem):
        newpinned = self.getSkeletonContext().getObject().notPinnedNodes()
        self.pinnednodes.start()
        self.pinnednodes.clear()
        self.pinnednodes.pin(newpinned)
        self.pinnednodes.signal()

    def unpinall(self, menuitem):
        if self.pinnednodes is not None:
            self.pinnednodes.start()
            self.pinnednodes.clear()
            self.pinnednodes.signal()

    def undoPin(self, menuitem):
        if self.pinnednodes is not None:
            self.pinnednodes.undo()
            self.pinnednodes.signal()

    def redoPin(self, menuitem):
        if self.pinnednodes is not None:
            self.pinnednodes.redo()
            self.pinnednodes.signal()
            
    # switchboard "new pinned nodes"
    def newPinnedNodes(self, pinnednodesSelection, skelcontext):
        switchboard.notify((self.gfxwindow(), "new nodes pinned"),
                           pinnednodesSelection)
        switchboard.notify("redraw")

    tip="Pin nodes."

    discussion="""<para>
    Pinned nodes are immobile during &skel; modifications.  This menu
    contains tools for pinning and unpinning nodes with mouse input.
    </para>"""

toolbox.registerToolboxClass(PinnedNodesToolbox, ordering=2.7)
