// -*- C++ -*-
// $RCSfile: rank3tensor.C,v $
// $Revision: 1.7 $
// $Author: langer $
// $Date: 2005/05/13 18:44:06 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */
#include <oofconfig.h>

#include "common/ooferror.h"
#include "engine/fieldindex.h"
#include "engine/matvec.h"
#include "engine/ooferror.h"
#include "engine/symeig3.h"
#include "engine/rank3tensor.h"
#include "engine/property/elasticity/cijkl.h"
#include <iomanip>
#include <iostream>
using namespace std;

Rank3Tensor::Rank3Tensor(const Rank3Tensor& sm) :  m0(3), m1(3), m2(3), nrows(3) {
  (*this)=sm;
}
    
Rank3Tensor::~Rank3Tensor() {

}



Rank3Tensor& Rank3Tensor::operator=(const Rank3Tensor& sm) {
  if(&sm != this) {
    m0 = sm.m0;
    m1 = sm.m1;
    m2 = sm.m2;
  }
  
  return *this;
}

// access

double Rank3Tensor::operator()(int i, int j, int k) const {
#ifdef DEBUG
  if(i >= nrows || j >= nrows || k>=nrows || i<0 || j<0 || k<0) {
    std::cerr << "Rank3Tensor error: i=" << i << " j=" << j
	      << " k=" << k
	      << " nrows=" << nrows << std::endl;
    abort();
  }
#endif
  if(i==0)
    {
      return m0(j,k);
    }  
  if(i==1)
    {
      return m1(j,k);
    }  
  return m2(j,k);
}

double &Rank3Tensor::operator()(int i, int j, int k) {
#ifdef DEBUG
  if(i >= nrows || j >= nrows || k>=nrows || i<0 || j<0 || k<0) {
    std::cerr << "Rank3Tensor error: i=" << i << " j=" << j
	      << " k=" << k
	      << " nrows=" << nrows<< std::endl;
    abort();
  }
#endif
  if(i==0)
    {
      return m0(j,k);
    }  
  if(i==1)
    {
      return m1(j,k);
    }  
  return m2(j,k);
}

double &Rank3Tensor::operator()(const int& i, const SymTensorIndex &jk) {
  return operator()(i, jk.row(), jk.col());
}

double Rank3Tensor::operator()(const int& i, const SymTensorIndex &jk) const
{
  return operator()(i, jk.row(), jk.col());
}


Rank3Tensor &Rank3Tensor::operator+=(const Rank3Tensor &a) {
  m0 +=a.m0;
  m1 +=a.m1;
  m2 +=a.m2;
  return *this;
}

Rank3Tensor &Rank3Tensor::operator-=(const Rank3Tensor &a) {
  m0 -= a.m0;
  m1 -= a.m1;
  m2 -= a.m2;
  return *this;
}

SymmMatrix operator*(const Rank3Tensor &a, const VECTOR_D &x) {
  int nrows = a.nrows;
  if(x.size() != nrows) 
    throw ErrProgrammingError("Vector has wrong dimension",
			      __FILE__, __LINE__);
  if(x.size() != nrows) abort();
  SymmMatrix result(nrows);
  for(int i=0; i<nrows; i++) {
    for(int j=0; j<nrows; j++)
      for(int k=0; k<nrows;k++)
	result(i,j) += a(k,i,j)*x(k);
  }
  return result;
}

// Compute A^ T (*this) A
Rank3Tensor Rank3Tensor::transform(const MATRIX_D &A) const {
  if(A.nrows() != nrows) {
    std::cerr << " A.nrows()" << A.nrows() << " nrows" << nrows << std::endl;
    throw ErrProgrammingError("Matrix has wrong dimension",
			      __FILE__, __LINE__);
  }
  //from Nye page 111, equation 5
  Rank3Tensor result;
  for(int i=0; i<nrows; i++) {
    for(int j=0; j< nrows; j++) {
      for (int k=0; k<nrows; k++) {
	double &r = result(i,j, k);
	  for(int l=0; l<nrows; l++)
	    for(int m=0; m<nrows; m++)
	      for(int n=0; n<nrows; n++)
	    r += A(i, l) * A(j, m) * A(k,n) * (*this)(l, m, n);
      }
    }
  }
  return result;
}

Rank3Tensor operator*(double x, const Rank3Tensor &A) {
  Rank3Tensor result(A);
  result.m0 *= x;
  result.m1 *= x;
  result.m2 *= x;
  return result;
}

Rank3Tensor operator*(const Rank3Tensor &A, double x) {
  Rank3Tensor result(A);
  result.m0 *= x;
  result.m1 *= x;
  result.m2 *= x;
  return result;
}


Rank3Tensor &Rank3Tensor::operator*=(double x) {
  m0 *= x;
  m1 *= x;
  m2 *= x;
  return *this;
}

Rank3Tensor &Rank3Tensor::operator/=(double x) {
  return operator*=(1./x);
}

Rank3Tensor operator/(const Rank3Tensor &A, double x) {
  Rank3Tensor result(A);
  result *= 1./x;
  return result;
}



std::string Rank3Tensor::classname_("Rank3Tensor");
std::string Rank3Tensor::modulename_("oof2.SWIG.engine.rank3tensor");



SymmMatrix &Rank3Tensor::operator()(const int i) {
#ifdef DEBUG
  if(i >= nrows || i<0) {
    std::cerr << "Rank3Tensor error: i=" 
	      << " nrows=" << nrows<< std::endl;
    abort();
  }
#endif
  if(i==0)
    {
      return m0;
    }  
  if(i==1)
    {
      return m1;
    }  
  return m2;
}

SymmMatrix Rank3Tensor::operator()(const int i) const {
#ifdef DEBUG
  if(i >= nrows || i<0) {
    std::cerr << "Rank3Tensor error: i=" 
	      << " nrows=" << nrows<< std::endl;
    abort();
  }
#endif
  if(i==0)
    {
      return m0;
    }  
  if(i==1)
    {
      return m1;
    }  
  return m2;
}

Rank3Tensor Rank3Tensor::operator*(const Cijkl& c) 
{

  Rank3Tensor e;
  for(int i=0; i<3;i++)
    for(int j=0; j<3;j++)
      for(int k=0; k<3;k++)
	for(int l=0; l<3;l++)
	  for(int m=0; m<3;m++)
	    e(i,j,k)+=(*this)(i,l,m)*c(l,m,j,k);
  return e;
}


SymmMatrix Rank3Tensor::operator*(const VECTOR_D& E) {

  SymmMatrix result(nrows);
  
  //Perform multiplications
  for(int i=0;i<nrows; i++)
    for(int j=0; j<nrows;j++)
      for(int k=0; k<nrows;k++)
	  result(i,j) += (*this)(k,i,j)*E(k);

return result;
}


Rank3Tensor operator*(const Cijkl &a, const VECTOR_D &x) {
  //if(a.nrows() != x.size()) {
  //  throw ErrProgrammingError("Cijkl object and vector have wrong dimension",
  //			      __FILE__, __LINE__);
  // }
  Rank3Tensor result;
  int nrows = x.size();
  for(int i=0; i<nrows; i++) 
    for(int j=0; j<nrows; j++)
      for(int k=0; k<=j; k++)
	{
	  double &r = result(i,j,k);
	  for(int l=0; l<nrows; l++)
	    r += a(l,i,j,k)*x(l);
	}
  return result;
}

Rank3Tensor operator*(const Cijkl& c, const Rank3Tensor& d) 
{

  Rank3Tensor e;
  int nrows = d.nrows;
  for(int i=0; i<nrows;i++)
    for(int j=0; j<nrows;j++)
      for(int k=0; k<nrows;k++)
	for(int l=0; l<nrows;l++)
	  for(int m=0; m<nrows;m++)
	      e(i,j,k)+=d(i,l,m)*c(l,m,j,k);
  return e;
}


std::ostream &operator<<(ostream &os, const Rank3Tensor &dd) {
  os.setf(std::ios::scientific, std::ios::floatfield);
  os << "[ ";
  for(int i=0; i<dd.nrows; i++){
    for(int j=0;j<dd.nrows;j++) {
      for(int k=0; k<dd.nrows; k++)
	os << dd(i, j, k) << " ";
      os << endl;
    }
    os << endl;
  }
   os << "]";
  return os ;
}
