// -*- C++ -*-
// $RCSfile: oofimage.C,v $
// $Revision: 1.62 $
// $Author: langer $
// $Date: 2005/06/08 15:06:36 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include <Python.h>
#include "image/oofimage.h"
#ifdef HAVE_MPI
#include "common/mpitools.h"
#endif ////HAVE_MPI
#include "common/IO/bitoverlay.h"
#include "common/IO/stringimage.h"
#include "common/ooferror.h"
#include "common/boolarray.h"
#include "common/doublearray.h"
#include <math.h>
#include <set>

OOFImage::OOFImage(const std::string &name, const std::string &filename)
  : name_(name), stringImage_(0)
{
  // It is apparently necessary to construct a null image first and
  // then perform a "read" operation on it, to get the data from a file.
  // The more obvious "image=Magick::Image(filename)" does not throw
  // an exception if the file cannot be found, it just returns a 
  // mangled and useless Magick::Image object. 
  image = Magick::Image();
  try {
    image.read(filename);
  }
  catch (Magick::Exception &error) {
    // Magick::Exceptions have to be converted into OOF2
    // ImageMagickErrors so that they'll be handled properly by the
    // SWIG exception typemap.
    throw ImageMagickError(error.what());
  }
  image.flip();		// real coordinates don't start at the top
  setup();
}

OOFImage::OOFImage(const std::string &name)
  : name_(name), stringImage_(0)
{
}

// Constructor for making a new blank image the same size 
// and size-in-pixels as an old one.  You can't just pass in 
// the old image, because that fuction signature is already 
// taken by the real copy constructor.
OOFImage::OOFImage(const std::string &name, const Coord &size, 
		   const Magick::Geometry &g) 
  : name_(name), size_(size), stringImage_(0), 
    image(g, Magick::Color(0,0,0))
{
  setup();
}

// Copy constructor that makes an explicit copy -- this bypasses a
// cache-entry-counting bug in the Magick++ API's "Image" class copy
// constructor.  This bug is to have been corrected in 5.5.7-16 and
// later. 
OOFImage::OOFImage(const OOFImage &other) 
  : image(other.geometry(), Magick::Color(0,0,0)),
    name_(other.name_),
    size_(other.size_),
    stringImage_(0)
{
  
  Magick::Geometry g = geometry();
  int w = g.width();
  int h = g.height();

  // Fast PixelPacket copy -- uses a lot of memory for big images.
  const Magick::PixelPacket *orig = other.image.getConstPixels(0,0,w,h);
  Magick::PixelPacket *copy = image.setPixels(0,0,w,h);
  
  for(int i=0;i<h*w;i++) {
    copy[i]=orig[i];
  } 
  image.syncPixels();

    
  // Slow pixelColor way.  Uses less memory?
  // image.modifyImage();
  // for(int r=0;r<g.height();r++)
  //   for(int c=0;c<g.width();c++) {
  //     image.pixelColor(r,c,other.image.pixelColor(r,c));
  //  }

  setup();
}

OOFImage *newImageFromData(const std::string &name, const ICoord *isize,
			   const std::vector<unsigned short> *data)
{
  return new OOFImage(name, *isize, "RGB", Magick::ShortPixel, &((*data)[0]));
}

OOFImage::OOFImage(const std::string &name, const ICoord &isize,
		   const std::string &map,
		   const Magick::StorageType storage,
		   const void *data) 
  : name_(name),
    image(isize(0), isize(1), map, storage, data),
    stringImage_(0)
{
  setup();
}


void OOFImage::setup() {
  try {
    Magick::Geometry sighs = image.size();
    sizeInPixels_ = ICoord(sighs.width(), sighs.height());
  }
  catch (Magick::Exception &e) {
    throw ImageMagickError(e.what());
  }
  
  // Scale factor for converting ImageMagick rgb values to floats in
  // [0,1].  Using "Magick::MaxRGB" doesn't work, so we have to use
  // "using namespace Magick" to get access to MaxRGB.
  using namespace Magick;
  scale = 1./MaxRGB;
}

OOFImage::~OOFImage() {
  if(stringImage_) delete stringImage_;
}

void OOFImage::save(const std::string &filename) {
  Magick::Image copy = image;
  try {
    copy.flip();			// undo flip in constructor
    copy.write(filename);
  }
  catch (Magick::Exception &e) {
    throw ImageMagickError(e.what());
   }
}

void OOFImage::setSize(const Coord *sighs) {
  size_ = *sighs;
}

ICoord OOFImage::pixelFromPoint(const Coord *point) const {
  double xx = (*point)(0)/size_(0)*sizeInPixels_(0);
  double yy = (*point)(1)/size_(1)*sizeInPixels_(1);
  if (xx == sizeInPixels_(0))
    xx = sizeInPixels_(0) - 1.0;
  if (yy == sizeInPixels_(1))
    yy = sizeInPixels_(1) - 1.0;
  return ICoord((int) floor(xx), (int) floor(yy));
}

OOFImage *OOFImage::clone(const std::string &nm) const {
  OOFImage *copy = new OOFImage(nm);
  try {
    copy->image = Magick::Image(image); // Magick::Image copy constructor
  }
  catch (Magick::Exception &e) {
    throw ImageMagickError(e.what());
  }
  copy->setup();
  copy->size_ = size_;
  copy->setMicrostructure(microstructure);
  return copy;
}

ShapedOOFImage *OOFImage::makeShapedImage(const BitmapOverlay *bitmap,
					  const double opacity) const
{
  ShapedOOFImage *soi = new ShapedOOFImage(name(), *this, *bitmap,
						 opacity);
  return soi;
}

std::string OOFImage::comment() const {
  return image.comment();
}

void OOFImage::imageChanged() {
  ++timestamp;			// marks image as changed
}

StringImage *OOFImage::stringimage() const {
  // Convert image into a string suitable for constructing a GdkImlibImage.
  if(stringImage_) delete stringImage_;
  stringImage_ = new StringImage(&sizeInPixels_, &size_);
  for(int j=0; j<sizeInPixels_(1); j++) {
    for(int i=0; i<sizeInPixels_(0); i++) {
      ICoord where(i,j);
      CColor color = (*this)[where];
      stringImage_->set(&where, &color);
    }
  }
  return stringImage_;
}

std::vector<unsigned short> *OOFImage::getPixels() {
  // Magick::Image::write isn't const, so this function isn't const either.
  int n = 3*sizeInPixels_(0)*sizeInPixels_(1);
  std::vector<unsigned short> *pxls = new std::vector<unsigned short>(n);
  image.write(0, 0, sizeInPixels_(0), sizeInPixels_(1),
	      "RGB", Magick::ShortPixel, &(*pxls)[0]);
  return pxls;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Access to individual pixels.

const CColor OOFImage::operator[](const ICoord &c) const {
  try {
    Magick::Color color = image.pixelColor(c(0), c(1));
    return CColor(color.redQuantum()*scale,
		  color.greenQuantum()*scale,
		  color.blueQuantum()*scale);
  }
  catch (Magick::Exception &e) {
    throw ImageMagickError(e.what());
  }
}

// TODO LATER: It may be useful to have "block-set" routines which use
// the ImageMagick PixelPacket routine to set many pixels together.
// This is alleged to be much faster.  See the OOFImage copy
// constructor definition for an example.
void OOFImage::set(const ICoord &c, const CColor &color) {
  try {
    Magick::ColorRGB culler(color.getRed(), color.getGreen(), color.getBlue());
    image.pixelColor(c(0), c(1), culler);
  }
  catch (Magick::Exception &e) {
    throw ImageMagickError(e.what());
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Conversion to and from arrays of double, int, or bool.  This is not
// an efficient way of getting or setting just a few pixels.  These
// routines all take as an argument a function f which converts from
// CColor to the appropriate type (double, int, or bool) or vice
// versa.

// TODO LATER: Is it better to make these functions use
// OOFImage::image::write, like OOFImage::getPixels does?  Doing so
// would require allocating another array, so it's not obviously more
// efficient.

Array<double> OOFImage::convert(double (*f)(const CColor&)) const {
  Array<double> arr(sizeInPixels_(0), sizeInPixels_(1));
  for(Array<double>::iterator i=arr.begin(); i!=arr.end(); ++i) {
    arr[i] = (*f)((*this)[i.coord()]);
  }
  return arr;
}

Array<int> OOFImage::convert(int (*f)(const CColor&)) const {
  Array<int> arr(sizeInPixels_(0), sizeInPixels_(1));
  for(Array<int>::iterator i=arr.begin(); i!=arr.end(); ++i) {
    arr[i] = (*f)((*this)[i.coord()]);
  }
  return arr;
}

Array<bool> OOFImage::convert(bool (*f)(const CColor&)) const {
  Array<bool> arr(sizeInPixels_(0), sizeInPixels_(1));
  for(Array<bool>::iterator i=arr.begin(); i!=arr.end(); ++i) {
    arr[i] = (*f)((*this)[i.coord()]);
  }
  return arr;
}

void OOFImage::set(const Array<double> &array, CColor (*f)(double)) {
  for(Array<double>::const_iterator i=array.begin(); i!=array.end(); ++i)
    set(i.coord(), (*f)(array[i]));
  imageChanged();
}

void OOFImage::set(const Array<int> &array, CColor (*f)(int)) {
  for(Array<int>::const_iterator i=array.begin(); i!=array.end(); ++i)
    set(i.coord(), (*f)(array[i]));
  imageChanged();
}

void OOFImage::set(const Array<bool> &array, CColor (*f)(bool)) {
  for(Array<bool>::const_iterator i=array.begin(); i!=array.end(); ++i)
    set(i.coord(), (*f)(array[i]));
  imageChanged();
}
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// std::vector<ICoord>* OOFImage::getColorPoints(const CColor &ref, 
// 					      const ColorDifference &diff)
// const
// {
//   std::vector<ICoord>* icoordlist = new std::vector<ICoord>(0);
//   // Be greedy, reserve enough space for all the pixels.
//   icoordlist->reserve(sizeInPixels_(0)*sizeInPixels_(1));
//   for(ConstOOFImageIterator i=this->begin(); i!=this->end(); ++i) {
//     if ( diff.contains(ref, *i) )
//       icoordlist->push_back(i.coord());
//   }
//   return icoordlist;  // memory leak
// }

void OOFImage::getColorPoints(const CColor &ref, 
			      const ColorDifference &diff,
			      BoolArray &selected)
const
{
  for(ConstOOFImageIterator i=this->begin(); i!=this->end(); ++i)
    if (diff.contains(ref, *i)) selected[i.coord()] = true;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

OOFImage::iterator OOFImage::begin() {
  return OOFImageIterator(*this, 0);
}

OOFImage::const_iterator OOFImage::begin() const {
  return ConstOOFImageIterator(*this, 0);
}

OOFImage::iterator OOFImage::end() {
  return OOFImageIterator(*this, sizeInPixels_(0)*sizeInPixels_(1));
}

OOFImage::const_iterator OOFImage::end() const {
  return ConstOOFImageIterator(*this, sizeInPixels_(0)*sizeInPixels_(1));
}


ICoord OOFImageIterator::coord() const {
  int width = image.sizeInPixels()(0);
  int y = pos/width;
  int x = pos - width*y;
  return ICoord(x, y);
}

ICoord ConstOOFImageIterator::coord() const {
  int width = image.sizeInPixels()(0);
  int y = pos/width;
  int x = pos - width*y;
  return ICoord(x, y);
}

bool operator==(const OOFImageIterator &a, const OOFImageIterator &b) {
  return a.pos == b.pos;
}

bool operator!=(const OOFImageIterator &a, const OOFImageIterator &b) {
  return a.pos != b.pos;
}


bool operator==(const ConstOOFImageIterator &a, const ConstOOFImageIterator &b){
  return a.pos == b.pos;
}

bool operator!=(const ConstOOFImageIterator &a, const ConstOOFImageIterator &b){
  return a.pos != b.pos;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Examples of image modification routines.

void OOFImage::flip(const std::string &axis) {
  if(axis == "x")
    image.flop();		// ImageMagick function call
  else if(axis == "y")
    image.flip();
  else if(axis == "xy") {
    image.flip();
    image.flop();
  }
  imageChanged(); // call this after using ImageMagick modification routines
}

void OOFImage::dim(double factor) {
  for(OOFImage::iterator i=begin(); i!=end(); ++i) {
    CColor c = *i;
    c.dim(factor);
    set(i.coord(), c);
  }
  imageChanged();		// call this after setting pixels directly
}

double color2gray(const CColor &color) {
  return color.getGray();
}

CColor gray2color(double x) {
  return CColor(x, x, x);
}

CColor bool2color(bool x) {
  if(x)
    return CColor(1,1,1);
  else
    return CColor(0,0,0);
}

void OOFImage::gray() {
  // simple example using OOFImage::convert()
  Array<double> arr = convert(color2gray);
  set(arr, gray2color);		// calls imageChanged() itself
}

void OOFImage::fade(double factor) {
  for(OOFImage::iterator i=begin(); i!=end(); ++i) {
    CColor c = *i;
    c.fade(factor);
    set(i.coord(), c);
  }
  imageChanged();
}

void OOFImage::blur(double radius, double sigma) {
  image.blur(radius, sigma);
  imageChanged();
}

void OOFImage::charcoal(double radius, double sigma) {
  image.charcoal(radius, sigma);
  imageChanged();
}

void OOFImage::contrast(bool sharpen) {
  image.contrast(sharpen);
  imageChanged();
}

void OOFImage::despeckle() {
  image.despeckle();
  imageChanged();
}

void OOFImage::edge(double radius) {
  image.edge(radius);
  imageChanged();
}

void OOFImage::enhance() {
  image.enhance();
  imageChanged();
}

void OOFImage::equalize() {
  image.equalize();
  imageChanged();
}

void OOFImage::medianFilter(double radius) {
  image.medianFilter(radius);
  imageChanged();
}

void OOFImage::negate(bool grayscale) {
  image.negate(grayscale);
  imageChanged();
}

void OOFImage::normalize() {
  image.normalize();
  imageChanged();
}

void OOFImage::reduceNoise(double radius) {
  image.reduceNoise(radius);
  imageChanged();
}

void OOFImage::sharpen(double radius, double sigma) {
  image.sharpen(radius, sigma);
  imageChanged();
}

BoolArray threshold(const DoubleArray& image1, double tolerance)
{
// This function converts a grayscale image to binary (black/white).
// Any value above the tolerance value (passed in) is changed to white
// and any value under the tolerance is change to black.

  bool color;  // The color of the current pixel
  int w = image1.width();  // Image width
  int h = image1.height();  // Image height
  BoolArray newimage(w, h);  // Copies the image from the passed in (image1) to the function's image (image)
  for (DoubleArray::const_iterator i = image1.begin(); i != image1.end(); ++i) {  // For width
    if (image1[i]<=tolerance) // If cell is under tolerance,
      color = false;  // Make low pixels black
    else
      color = true; // Make high pixels white
    newimage[i.coord()] = color;  // Cell takes new color as set above	
  }

  return newimage;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ShapedOOFImage::ShapedOOFImage(const std::string &name,
				     const OOFImage &srcimage,
				     const BoolArray &bits)
  : name_(name), size_(srcimage.size()), isize_(srcimage.sizeInPixels())
{
  const StringImage *srcstring = srcimage.stringimage();
  stringimage_ = srcstring->mask(bits);
}

ShapedOOFImage::ShapedOOFImage(const std::string &name,
				     const OOFImage &srcimage,
				     const BitmapOverlay &bitmap,
				     double opacity)
  : name_(name), size_(srcimage.size()), isize_(srcimage.sizeInPixels())
{
  const StringImage *srcstring = srcimage.stringimage();
  stringimage_ = srcstring->blend(&bitmap, opacity);
}

ShapedOOFImage::~ShapedOOFImage() {
  delete stringimage_;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ImageMagickError::ImageMagickError(const std::string &messg) {
  // Make sure that all quotation marks in messg are escaped.
  for(int i=0; i<messg.size(); ++i) {
    char c = messg[i];
    if(c == '"' || c == '\'')
      msg += '\\';
    msg += c;
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

#ifdef HAVE_MPI
void send_image(OOFImage *img, const int origin)
{

  // function is called by both front end
  // front end deconstructs the image into simple 
  // pieces and broadcasts it. Back-end reads the pieces
  // and reconstructs it.
  int rnk = rank();
  if (rnk == origin)
    {
      //for(OOFImage::iterator c = img->begin(); c != img->end(); ++c)
      //std::cerr << "color_out=" << *c << std::endl;
      //deconstruct image
      std::vector<unsigned short> * short_int_img = img->getPixels();
      std::string nnaammee = img->name();
      Coord physize = img->size();
      double physize_x = physize(0);
      double physize_y = physize(1);
      ICoord pixsize = img->sizeInPixels();
      int pixsize_x = pixsize(0);
      int pixsize_y = pixsize(1);
      //send elementary components
      broadcast_string(nnaammee, origin); // note that rnk == origin
      //std::cerr << "before sending physical size" << std::endl;
      broadcast_double(physize_x, origin);
      broadcast_double(physize_y, origin);
      //std::cerr << "before sending pixel size" << std::endl;
      broadcast_int(pixsize_x, origin);
      broadcast_int(pixsize_y, origin);
      //std::cerr << "before sending vector of shorts" << std::endl;
      broadcast_vector_of_shorts(short_int_img, origin);
      //std::cerr << "after sending vector of shorts" << std::endl;
      delete short_int_img;
    }
  else
    {
      std::cerr << "sender and rank do not coincide!!! rank= " << rnk << " sender = " << origin << std::endl;
      abort();
    }

}

OOFImage * receive_image(const int source)
{
  if(rank() != source)
    {
      //receive the different parts of the image from the front end
      std::string nnaammee = receive_broadcast_string(source);
      //std::cerr << "before receiving physical size" << std::endl;
      double physize_x = receive_broadcast_double(source);
      double physize_y = receive_broadcast_double(source);
      //std::cerr << "before allocating Coord to physical size" << std::endl;
      Coord * physize = new Coord(physize_x, physize_y);
      //std::cerr << "before receiving pixel size" << std::endl;
      int  pixsize_x = receive_broadcast_int(source);
      int pixsize_y = receive_broadcast_int(source);
      ICoord pixsize(pixsize_x, pixsize_y);
      //std::cerr << "before recieving vector of shorts" << std::endl;
      std::vector<unsigned short> * temp  = receive_broadcast_vector_of_shorts(source);
      //std::cerr << "after receiving vector of shorts" << std::endl;
      //assemble copy on back-end.
      OOFImage * img = newImageFromData(nnaammee, &pixsize, temp);
      //for(OOFImage::iterator c = img->begin(); c != img->end(); ++c)
      //std::cerr << "color_in=" << *c << std::endl;
      //std::cerr << "after creating image clone" << std::endl;
      img->setSize(physize); //<-- this is the back-end image
      //std::cerr << "after setting physical size" << std::endl;
      delete temp;
      //std::cerr << "before returning " << std::endl;
      //return received image
      return img; // back end returns received img
    }
  std::cerr << "deadlock!! Source is also waiting to receive the same message!!!" << source << std::endl;
  abort();
  return 0;
}
// end parallelize_image

#endif //HAVE_MPI
